#include <iostream> // cin, cout
#include <cmath> // fabs
#include <GL/glut.h>

using namespace std;

//------------------------------------------------------------
// variables globales y defaults

int 
  w=480,h=360, // tamao de la ventana
  npoints=0, // cantidad de puntos
  ep=-1; // indice del punto a editar

float // luces y colores en float
  fondo[]={0.85f,0.9f,0.95f,1.f},       // color de fondo
  line_color[]={.4f,.6f,.8f,.5f},
  point_color[]={.1f,.2f,.3f,.5f};

double
  pt[1000], // los puntos
  c[2],r; // circunferencia
  
bool 
  dibujar=true,      // false si esta minimizado
  cl_info=true;     // informa por la linea de comandos

short modifiers=0;  // ctrl, alt, shift (de GLUT)


//------------------------------------------------------------
// funciones globales

// utiles

template <class T> static inline int redondea(const T& value)
{return (value>0)? (int(value+.5)): (int(value-.5));}

template <class T> static inline void intercambia(T&t1, T&t2) //swap en conflicto con stl
{T t0=t1; t1=t2; t2=t0;} // cambia el contenido y no la direccion (lento pero seguro)

// mnimo y maximo
template <class T> static inline const T &minimo(const T &t1,const T &t2)
  {if (t1<t2) return t1; else return t2;}
template <class T> static inline const T &maximo(const T &t1,const T &t2)
  {if (t1>t2) return t1; else return t2;}

// potencia >0 (ojo que no testea negativos!!)
template <class T> static inline T pown(const T &inp, int p)
  {T ret=1;  while (p--) ret*=inp; return ret;} // ojo el 1
template <class T> static inline T cuadrado(const T &inp)
  {return inp*inp;}

// raiz cuadrada entera
// s=(s+x/s)>>1, empezando de x/2 da el mayor entero cuyo cuadrado es < x
// yo calculo raiz de 4x, si es impar redondeo a mas y si es par a menos
static int raiz(int x){
  if (x<=0) return 0;
  int s=x<<1,sant=s<<1;
  x=sant;
  while (s<sant) {
    sant=s; 
    s=(s+x/s)>>1;
  }
  s=sant; s>>=1;
  if ((s<<1)!=sant) s++;
  return s; 
}

//------------------------------------------------------------
// Circunferencia aproximante

bool invert_matrix(const double *m, double* inv, int dim=3){
  if (dim==3) {
    double det=
       m[0]*m[4]*m[8]+m[1]*m[5]*m[6]+m[2]*m[3]*m[7]-
      (m[6]*m[4]*m[2]+m[7]*m[5]*m[0]+m[8]*m[3]*m[1]);
    if (det==0) return false;
    inv[0]= (m[4]*m[8]-m[7]*m[5])/det;
    inv[1]=-(m[3]*m[8]-m[6]*m[5])/det;
    inv[2]= (m[3]*m[7]-m[6]*m[4])/det;
    inv[3]=-(m[1]*m[8]-m[7]*m[2])/det;
    inv[4]= (m[0]*m[8]-m[6]*m[2])/det;
    inv[5]=-(m[0]*m[7]-m[6]*m[1])/det;
    inv[6]= (m[1]*m[5]-m[4]*m[2])/det;
    inv[7]=-(m[0]*m[5]-m[3]*m[2])/det;
    inv[8]= (m[0]*m[4]-m[3]*m[1])/det;
    return true;
  }
  else return false;
}

/*
    a=2c,r^2-c^2 p=_p,1
    eqj: sumk(ak sumi(pij pik)) = sumi((pi^2-1) pij)
*/
bool fit_circle(int n,const double *_p,double *c,double &r,int d=2){
  int i,j,k,D=d+1;
  size_t szd=sizeof(double);
  double 
    *p=(double*) calloc(D*n,szd),
    *p2=(double*) calloc(n,szd),
    *sumpjpk=(double*) calloc(D*D,szd),
    *sump2pj=(double*) calloc(D,szd),
    *inv=(double*) calloc(D*D,szd),
    *soln=(double*) calloc(D,szd),
    c2=0;

  //p,1
  for (i=0;i<n;i++){
    for (j=0;j<d;j++) {
      p[D*i+j]=_p[d*i+j];
      p2[i]+=_p[d*i+j]*_p[d*i+j];
    }
    p[D*i+j]=1;
  }

  // sumas
  for (i=0;i<n;i++){
    for(j=0;j<D;j++){
      for (k=0;k<D;k++){
        sumpjpk[D*j+k]+=p[D*i+j]*p[D*i+k];
      }
      sump2pj[j]+=p2[i]*p[D*i+j];
    }
  }  

  // solucion
  if (!invert_matrix(sumpjpk,inv,d+1)) return false;
  for(j=0;j<D;j++) {
    for(k=0;k<D;k++){ 
      soln[j]+=inv[D*j+k]*sump2pj[k];
    }
    if (j<d) {c[j]=soln[j]/2; c2+=c[j]*c[j];} 
    else r=sqrt(soln[j]+c2);
  }

  free (p); free(p2);
  free(sumpjpk); free(sump2pj);
  free(inv); free(soln);

  return true;
}


//------------------------------------------------------------
// RASTERIZACION


void circulo_Bresenham(int xc, int yc, int r) {  
  glBegin(GL_POINTS);
  if (!r){
    glVertex2i(xc,yc);
    glEnd();
    return;
  }

  int x=0, y=r, h=1-r, E=3, SE=5-(r<<1);
  while (y>x){
    glVertex2i( x+xc, y+yc);
    glVertex2i( x+xc,-y+yc);
    glVertex2i(-x+xc, y+yc);
    glVertex2i(-x+xc,-y+yc);
    
    glVertex2i( y+xc, x+yc);
    glVertex2i( y+xc,-x+yc);
    glVertex2i(-y+xc, x+yc);
    glVertex2i(-y+xc,-x+yc);
//cout << x << "," << y << "\n";
    if (h>=0){ // SE
      h+=SE; E+=2; SE+=4;
      y--;
    }
    else { // E
      h+=E; E+=2; SE+=2;
    }
    x++;
  }
  // x==y => no hace falta intercambiar
  glVertex2i( x+xc, y+yc);
  glVertex2i( x+xc,-y+yc);
  glVertex2i(-x+xc, y+yc);
  glVertex2i(-x+xc,-y+yc);
  glEnd();
//cout << endl;
}
void circulo_Bresenham(int xc, int yc, float r) {
  circulo_Bresenham(xc,yc,redondea(r));
}
void circulo_Bresenham(int xc, int yc, int x1, int y1) {
  circulo_Bresenham(xc,yc,raiz(cuadrado(x1-xc)+cuadrado(y1-yc)));
}
void circulo_Bresenham(float xc, float yc, float r) {
  circulo_Bresenham(redondea(xc),redondea(yc),redondea(r));
}
void circulo_Bresenham(float xc, float yc, float x1, float y1) {
  circulo_Bresenham(redondea(xc),redondea(yc),raiz(redondea(cuadrado(x1-xc)+cuadrado(y1-yc))));
}

// llamadas a los algoritmos
void tic(int x, int y, bool borrar=false){
  glColor4fv((borrar)? fondo : point_color);
  glPointSize(3.0); // tic con punto grueso
  glBegin(GL_POINTS);
    glVertex2i(x  ,y  );
/*    glVertex2i(x-1,y  ); // tic con crucesita
    glVertex2i(x  ,y-1);
    glVertex2i(x+1,y  );
    glVertex2i(x  ,y-1);*/
  glEnd();
  glPointSize(1.0);
}

void redibuja(bool borrar=false){
  if (!npoints) return;
  for(int i=0;i<npoints;i++){
    tic(pt[2*i],pt[2*i+1],borrar);
  }
  if (npoints>2){
    circulo_Bresenham(redondea(c[0]),redondea(c[1]),redondea(r));
  }
}

void borra(){redibuja(true);}

//============================================================
// callbacks

//------------------------------------------------------------
// redibuja los objetos
// Cada vez que hace un redisplay
void Display_cb() { // Este tiene que estar
  if (!dibujar) return;

  // borra el buffer de pantalla
  glClear(GL_COLOR_BUFFER_BIT); // o redibuja(true) ??
  redibuja();
  glutSwapBuffers();


#ifdef _DEBUG
  int errornum=glGetError();
  if (cl_info){
    while(errornum!=GL_NO_ERROR){
           if(errornum==GL_INVALID_ENUM) 
             cout << "GL_INVALID_ENUM" << endl;
      else if (errornum==GL_INVALID_OPERATION) 
             cout << "GL_INVALID_OPERATION" << endl;
      else if (errornum==GL_STACK_OVERFLOW) 
             cout << "GL_STACK_OVERFLOW" << endl;
      else if (errornum==GL_STACK_UNDERFLOW) 
             cout << "GL_STACK_UNDERFLOW" << endl;
      else if (errornum==GL_OUT_OF_MEMORY) 
             cout << "GL_OUT_OF_MEMORY" << endl;
      errornum=glGetError();
    }
  }
#endif // _DEBUG
}

//------------------------------------------------------------
// Maneja cambios de ancho y alto del screen
void Reshape_cb(int width, int height){
  if (cl_info) cout << "reshape " << width << "x" << height << endl;

  h=height; w=width;

  if (!h||!w) {dibujar=false; return;}
  dibujar=true;

  glViewport(0,0,w,h); // region donde se dibuja

	// matriz de proyeccion
	glMatrixMode(GL_PROJECTION);	glLoadIdentity();
  glOrtho(0,w,0,h,-1,1);

  glutPostRedisplay(); // avisa que hay que redibujar
}


//------------------------------------------------------------
// Teclado y Mouse
/*
GLUT ACTIVE SHIFT //Set if the Shift modifier or Caps Lock is active.
GLUT ACTIVE CTRL //Set if the Ctrl modifier is active.
GLUT ACTIVE ALT //Set if the Alt modifier is active.
*/
inline short get_modifiers() {return modifiers=(short)glutGetModifiers();}

// Movimientos del mouse
void Motion_cb(int x, int y){ // drag
  y=h-y; // el 0 esta ariba
  redibuja(true); // borra    
  pt[2*ep+0]=x;pt[2*ep+1]=y;
  if (npoints>2) fit_circle(npoints,pt,c,r);
  redibuja(); glutSwapBuffers();
}

// Clicks del mouse
// GLUT LEFT BUTTON, GLUT MIDDLE BUTTON, or GLUT RIGHT BUTTON
// The state parameter is either GLUT UP or GLUT DOWN
// glutGetModifiers may be called to determine the state of modifier keys

void Mouse_cb(int button, int state, int x, int y){
  y=h-y; // el 0 esta ariba
  if (button==GLUT_LEFT_BUTTON){
    if (state==GLUT_DOWN) {
      ep=-1;
      for (int i=0;i<npoints;i++){
        int d=abs(x-pt[2*i+0])+abs(y-pt[2*i+1]); // metrica del rombo
        if (d>5) continue;
        // edita (5 pixeles tolerancia)
        glutMotionFunc(Motion_cb); // callback para los drags
        ep=i;
        return;
      }
      // agrega un punto
      redibuja(true); // borra    
      pt[2*npoints+0]=x; pt[2*npoints+1]=y; npoints++;
      if (npoints>2) fit_circle(npoints,pt,c,r);
      redibuja(); glutSwapBuffers();
    } // down
    else if (state==GLUT_UP)
      glutMotionFunc(0); // anula el callback para los drags
  } // left
}

// Maneja pulsaciones del teclado (ASCII keys)
// x,y posicion del mouse cuando se teclea
void Keyboard_cb(unsigned char key,int x=0,int y=0) {  
  if (key==27){ // escape => exit
    get_modifiers();
    if (!modifiers)
      exit(EXIT_SUCCESS);
  }
  else if (key==127){ // DEL
    redibuja(true); // borra    
    if (ep>=0&&ep!=npoints-1) {
      for(int i=ep;i<npoints-1;i++) {pt[2*i]=pt[2*i+2];pt[2*i+1]=pt[2*i+3];}
    }
    else ep=npoints-1;
    npoints--;
    if (npoints>2) fit_circle(npoints,pt,c,r);
    redibuja(); glutSwapBuffers();
    cout << "Borrado el punto " << ep << endl;
  }
}

// Special keys (non-ASCII)
/*
  GLUT KEY F[1,12] F[1,12] function key.
  GLUT KEY LEFT Left directional key.
  GLUT KEY UP Up directional key.
  GLUT KEY RIGHT Right directional key.
  GLUT KEY DOWN Down directional key.
  GLUT KEY PAGE UP Page up directional key.
  GLUT KEY PAGE DOWN Page down directional key.
  GLUT KEY HOME Home directional key.
  GLUT KEY END End directional key.
  GLUT KEY INSERT Inset directional key.
*/
// aca es int key
void Special_cb(int key,int xm=0,int ym=0) {
  if (key==GLUT_KEY_F4){ // alt+f4 => exit
    if (get_modifiers()==GLUT_ACTIVE_ALT)
      exit(EXIT_SUCCESS);
  }
}

int integerv(GLenum pname){
  int value;
  glGetIntegerv(pname,&value);
  return value;
}
#define _PRINT_INT_VALUE(pname) #pname << ": " << integerv(pname) <<endl
 
//------------------------------------------------------------
// Inicializa GLUT y OpenGL
void initialize() {
	// pide color RGB y double buffering
	glutInitDisplayMode(GLUT_RGB|GLUT_DOUBLE);

  glutInitWindowSize(w,h); glutInitWindowPosition(50,50);

  glutCreateWindow("Circunferencia aproximante"); // crea el main window
  
  //declara los callbacks
  //los que no se usan no se declaran
  glutDisplayFunc(Display_cb); // redisplays
  glutReshapeFunc(Reshape_cb);
  glutKeyboardFunc(Keyboard_cb);
  glutSpecialFunc(Special_cb);
  glutMouseFunc(Mouse_cb);
  
  // ========================
  // estado normal del OpenGL
  // ========================
  glClearColor(fondo[0],fondo[1],fondo[2],1);  // color de fondo
  glMatrixMode(GL_MODELVIEW); glLoadIdentity(); // constante

  // ========================
  // info
  if (cl_info)
    cout << "Vendor:         " << glGetString(GL_VENDOR) << endl 
         << "Renderer:       " << glGetString(GL_RENDERER) << endl 
         << "GL_Version:     " << glGetString(GL_VERSION) << endl 
         << "GL_Extensions:  " << glGetString(GL_EXTENSIONS) << endl
         << "GLU_Version:    " << gluGetString(GLU_VERSION) << endl 
         << "GLU_Extensions: " << gluGetString(GLU_EXTENSIONS) << endl
         << _PRINT_INT_VALUE(GL_DOUBLEBUFFER)
         << _PRINT_INT_VALUE(GL_STEREO)
         << _PRINT_INT_VALUE(GL_AUX_BUFFERS)
         << _PRINT_INT_VALUE(GL_RED_BITS)
         << _PRINT_INT_VALUE(GL_GREEN_BITS)
         << _PRINT_INT_VALUE(GL_BLUE_BITS)
         << _PRINT_INT_VALUE(GL_ALPHA_BITS)
         << _PRINT_INT_VALUE(GL_DEPTH_BITS)
         << _PRINT_INT_VALUE(GL_STENCIL_BITS)
         << _PRINT_INT_VALUE(GL_ACCUM_RED_BITS)
         << _PRINT_INT_VALUE(GL_ACCUM_GREEN_BITS)
         << _PRINT_INT_VALUE(GL_ACCUM_BLUE_BITS)
         << _PRINT_INT_VALUE(GL_ACCUM_ALPHA_BITS)
         ;
  // ========================
}

//------------------------------------------------------------
// main
int main(int argc,char** argv) {
	glutInit(&argc,argv);// inicializa glut
	initialize(); // condiciones iniciales de la ventana y OpenGL
	glutMainLoop(); // entra en loop de reconocimiento de eventos
  return 0; //solo para que el builder se quede tranquilo
}
